import utils.context as context
import utils.log as log
import helpers.recon as recon
from utils.utils import fix_url, fix_filepath, cookie_string_to_dict, get_full_response

from argparse import ArgumentParser
from requests import Session, get


# Arguments
parser = ArgumentParser(description='A Web Analyser')
parser.add_argument('-u', '--url', type=str, help='The URL')
parser.add_argument('-o', '--output', type=str, help='The Output File')
parser.add_argument('--user', '--user-agent', type=str, help='The User-Agent to use', default='requests')
parser.add_argument('-c', '--cookies', type=str, help='Any cookies you need')
args = parser.parse_args()


# Set the Context
#context.url = fix_url(args.url)
#context.file = fix_filepath(args.output)

context.session = Session()
context.session.headers.update({'User-Agent': args.user})

if args.cookies:
    context.session.cookies.update(cookie_string_to_dict(args.cookies))

'''
# Log it all
log.success(f'Analysing {context.url}')
log.success(f'Saving output to {context.file}')


# Execute the different modules
recon.execute()
'''


import utils.log as log
import utils.context as context

from utils.utils import grab, get_full_response
from utils.constans import url_regex, resource_regex, jwt_regex
from helpers.analysis import redirect
from utils.constants import user_agents_list

from bs4 import BeautifulSoup, Comment
from colorama import Fore
from re import findall
from base64 import standard_b64decode

def execute():
    '''The function that calls all others'''

    robots()
    sitemap()
    cookies()
    get_jwts()
    redirects()
    comments()
    urls()
    resources()
    user_agents()


def robots():
    text = grab('robots.txt')

    log.info(f'Robots:\n{text}')


def sitemap():
    text = grab('sitemap.xml')

    log.info(f'Sitemap:\n{text}')


def cookies():
    r = grab(text=False)

    cookies = r.cookies.get_dict()

    if len(cookies) == 0:
        log.fail('No cookies found')
        return


    log.info('Cookies:')

    for x in cookies:
        cookie_data = x
        cookie_data = cookie_data.ljust(10, ' ')
        cookie_data += cookies[x]

        log.info(cookie_data, indent=1)


def get_jwts():
    response = get_full_response(grab(text=False))

    jwts = findall(jwt_regex, response)

    if len(jwts) == 0:
        log.error('No JWTs')
        return
    
    log.success('JWTs found:')

    for jwt in jwts:
        log.success(jwt, indent=1)

        # The last section of a JWT is the signature
        for section in jwt.split('.')[:-1]:
            log.info(standard_b64decode(section), indent=2)


def redirects():
    r = grab(text=False)

    history = r.history

    if len(history) == 0:
        log.fail('No redirects')
        return

    log.success('Redirects:')

    for url in history:
        red = Fore.RED + str(url.status_code) + Fore.RESET
        red = red.ljust(20, ' ')
        red += url.url

        log.info(red, indent=1)

        redirect(url.url)


def comments():
    r = grab()

    soup = BeautifulSoup(r, 'html.parser')
    comments = soup.find_all(string=lambda text: isinstance(text, Comment))

    if len(comments) == 0:
        log.fail('No comments')
        return

    log.success('Comments:')

    for c in comments:
        log.info(c, indent=1)


def urls():
    urls = findall(url_regex, grab())
    
    if len(urls) == 0:
        log.fail('No URLs')
        return

    log.success('URLs:')

    for url in urls:
        log.info(url, indent=1)


def resources():
    resources = findall(resource_regex, grab())
    
    if len(resources) == 0:
        log.fail('No Resources')
        return

    log.success('Resources:')

    for res in resources:
        log.info(res[1], indent=1)


def user_agents():
    length = len(context.session.get(context.url).text)
    log.info(f'Standard Response Length: {length}')

    for agent in user_agents_list:
        r = context.session.get(context.url, headers={'User-Agent': agent})

        log.info(agent, indent=1)

        if len(r.text) != length:
            log.success(f'Response size is different: {len(r.text)}', indent=2)
        else:
            log.fail('Default length', indent=2)


url_regex = r'\w+\:\w+\.\w+\.\w+'
resource_regex = r'("|\'){1}(/[\w\/]+)("|\'){1}'
file_regex = r'\w+\.\w+'
jwt_regex = r'^[A-Za-z0-9-_=]+\.[A-Za-z0-9-_=]+\.?[A-Za-z0-9-_.+/=]*$'


# Grabbed from
# https://www.networkinghowtos.com/howto/common-user-agent-list/

user_agents_list = [
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.36',                                                          # Chrome

    'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:53.0) Gecko/20100101 Firefox/53.0',                                                                                               # Firefox

    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/51.0.2704.79 Safari/537.36 Edge/14.14393',                                             # Edge

    'Mozilla/4.0 (compatible; MSIE 6.0; Windows NT 5.1; SV1)',                                                                                                                      # IE 6
    'Mozilla/5.0 (Windows; U; MSIE 7.0; Windows NT 6.0; en-US)',                                                                                                                    # IE 7
    'Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 5.1; Trident/4.0; .NET CLR 1.1.4322; .NET CLR 2.0.50727; .NET CLR 3.0.4506.2152; .NET CLR 3.5.30729)',                           # IE 8
    'Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.0; Trident/5.0;  Trident/5.0)',                                                                                                # IE 9
    'Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 6.2; Trident/6.0; MDDCJS)',                                                                                                     # IE 10
    'Mozilla/5.0 (compatible, MSIE 11, Windows NT 6.3; Trident/7.0; rv:11.0) like Gecko',                                                                                           # IE 11

    'Mozilla/5.0 (iPad; CPU OS 8_4_1 like Mac OS X) AppleWebKit/600.1.4 (KHTML, like Gecko) Version/8.0 Mobile/12H321 Safari/600.1.4',                                              # iPad
    'Mozilla/5.0 (iPhone; CPU iPhone OS 10_3_1 like Mac OS X) AppleWebKit/603.1.30 (KHTML, like Gecko) Version/10.0 Mobile/14E304 Safari/602.1',                                    # iPhone

    'Mozilla/5.0 (compatible; Googlebot/2.1; +http://www.google.com/bot.html)',                                                                                                     # Googlebot
    'Mozilla/5.0 (compatible; bingbot/2.0; +http://www.bing.com/bingbot.htm)',                                                                                                      # Bing Bot
    
    'Mozilla/5.0 (Linux; Android 6.0.1; SAMSUNG SM-G570Y Build/MMB29K) AppleWebKit/537.36 (KHTML, like Gecko) SamsungBrowser/4.0 Chrome/44.0.2403.133 Mobile Safari/537.36',        # Samsumg Phone
    'Mozilla/5.0 (Linux; Android 5.0; SAMSUNG SM-N900 Build/LRX21V) AppleWebKit/537.36 (KHTML, like Gecko) SamsungBrowser/2.1 Chrome/34.0.1847.76 Mobile Safari/537.36',            # Samsung Galaxy Note 3
    'Mozilla/5.0 (Linux; Android 6.0.1; SAMSUNG SM-N910F Build/MMB29M) AppleWebKit/537.36 (KHTML, like Gecko) SamsungBrowser/4.0 Chrome/44.0.2403.133 Mobile Safari/537.36',        # Samsung Galaxy Note 4

    'Mozilla/5.0 (Linux; U; Android-4.0.3; en-us; Galaxy Nexus Build/IML74K) AppleWebKit/535.7 (KHTML, like Gecko) CrMo/16.0.912.75 Mobile Safari/535.7',                           # Google Nexus
    
    'Mozilla/5.0 (Linux; Android 7.0; HTC 10 Build/NRD90M) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.83 Mobile Safari/537.36',                                        # HTC

    'curl/7.72.0',                                                                                                                                                                   # cURL

    'Wget/1.20 (linux-gnu)',                                                                                                                                                         # Wget

    'Lynx/2.8.8pre.4 libwww-FM/2.14 SSL-MM/1.4.1 GNUTLS/2.12.23'                                                                                                                    # Lynx
]


import utils.context as context
import utils.log as log

from utils.constants import url_regex, resource_regex, file_regex

from requests import get, Request
from os import getcwd
from os.path import exists, join

accept_codes = [200, 301]

def grab(file: str='', text=True):
    r = context.session.get(context.url + '/' + file)

    if r.status_code not in accept_codes:
        if file != '':
            log.fail(f'Could not grab {file} - failed with status code {r.status_code}')
    
    if not r:
        return ''

    if text:
        return r.text
    
    return r


def fix_url(url: str):
    '''Fixes incomplete urls'''

    splits = url.split('.')
    
    # URL => 3
    # IP  => 4
    if not 3 <= len(splits) <= 4:
        raise ValueError('Not a valid URL or IP')


    # Assume https, but ":" could also define port
    if ':' not in splits[0]:
        url = 'https://' + url
    
    return url


def fix_filepath(file: str):
    '''Updates the file path to the current working directory'''

    if not file:
        return

    # If there is a / or \, it's probably the full path already
    if '/' in file or '\\' in file:
        return file

    full_path = join(getcwd(), file)

    if exists(full_path):
        raise EnvironmentError('The output file already exists!')
    
    return full_path


def cookie_string_to_dict(cookies: str):
    cookie_dict = {}
    
    cookies = cookies.split('; ')

    for c in cookies:
        name, value = c.split('=')

        cookie_dict[name] = value
    
    return cookie_dict


def get_full_response(resp: Request):
    return ''.join(f'{header.lower()}: {value}\r\n' for header, value in resp.headers.items()) + '\r\n' + resp.text


from colorama import init, Fore

import utils.context as context

init(autoreset=True)


def __log(text, symbol, colour, end='\n', indent=0):
    text = str(text)

    plain_text = '\t' * indent + f'[{symbol}] {text}'
    coloured_text = '\t' * indent + f'[{colour}{symbol}{Fore.RESET}] {text}'

    if context.file:
        with open(context.file, 'a') as f:
            f.write(plain_text.rstrip() + '\n')

    print(coloured_text, end=end)


def info(text, end='\n', indent=0):             __log(text, '*', Fore.BLUE, end=end, indent=indent)
def fail(text, end='\n', indent=0):             __log(text, '-', Fore.RED, end=end, indent=indent)
def success(text, end='\n', indent=0):          __log(text, '+', Fore.GREEN, end=end, indent=indent)



from utils.utils import url_regex, resource_regex, file_regex
import utils.log as log

from urllib.parse import unquote

def redirect(url):
    if '?' not in url:
        log.fail('Redirect contains no GET parameters', indent=2)
        return
    

    # URL Decode
    params = unquote(url.split('?')[1]).split('&')

    for p in params:
        name, value = p.split('=')

        if re.match(url_regex, value):
            log.success('Redirect appears to contain a URL - possible RFI or use of another URL scheme?', indent=2)
            continue
        
        if re.match(resource_regex, value):
            log.success('Redirect appears to contain a resource - possible LFI?', indent=2)
            continue

        if re.match(file_regex, value):
            log.success('Redirect appears to contain a filename - possible LFI?', indent=2)
            continue
